@@ -1,6 +1,6 @@
 /*
     This file is part of Android File Transfer For Linux.
-    Copyright (C) 2015-2018  Vladimir Menshakov
+    Copyright (C) 2015-2020  Vladimir Menshakov
 
     This library is free software; you can redistribute it and/or modify it
     under the terms of the GNU Lesser General Public License as published by
@@ -28,11 +28,17 @@
 #include <mtp/ptp/ObjectPropertyListParser.h>
 #include <mtp/log.h>
 #include <mtp/version.h>
+#include <mtp/mtpz/TrustedApp.h>
 
-#include <sstream>
+#include <mtp/metadata/Metadata.h>
+#include <mtp/metadata/Library.h>
+
+#include <algorithm>
 #include <set>
+#include <sstream>
 
 #include <string.h>
+#include <stdlib.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/ioctl.h>
@@ -50,15 +56,18 @@ namespace
 		const char *b = prefix.c_str();
 		return strncasecmp(a, b, prefix.size()) == 0;
 	}
+
+	mtp::u64 FromHex(const std::string & str)
+	{ return strtoll(str.c_str(), NULL, 16); }
 }
 
 
 namespace cli
 {
 
-	Session::Session(const mtp::DevicePtr &device, bool showPrompt):
-		_device(device),
-		_session(_device->OpenSession(1)),
+	Session::Session(const mtp::SessionPtr &session, bool showPrompt):
+		_session(session),
+		_trustedApp(mtp::TrustedApp::Create(_session, GetMtpzDataPath())),
 		_gdi(_session->GetDeviceInfo()),
 		_cs(mtp::Session::AllStorages),
 		_cd(mtp::Session::Root),
@@ -84,6 +93,8 @@ namespace cli
 			_terminalWidth = ts.ws_col;
 #endif
 		}
+		try { if (_trustedApp) _trustedApp->Authenticate(); }
+		catch(const std::exception & ex) { error("mtpz authentication failed: ", ex.what()); }
 
 		AddCommand("help", "shows this help",
 			make_function([this]() -> void { Help(); }));
@@ -107,6 +118,8 @@ namespace cli
 			make_function([this]() -> void { List(true, true); }));
 		AddCommand("lsext-r", "<path> lists objects in <path> [extended info, recursive]",
 			make_function([this](const Path &path) -> void { List(path, true, true); }));
+		AddCommand("ls-objects", "<object-format-hex>",
+			make_function([this](const std::string &format) -> void { ListObjects(format); }));
 
 		AddCommand("put", "put <file> <dir> uploads file to directory",
 			make_function([this](const LocalPath &path, const Path &dst) -> void { Put(path, dst); }));
@@ -138,7 +151,7 @@ namespace cli
 		AddCommand("pwd", "resolved current object directory",
 			make_function([this]() -> void { CurrentDirectory(); }));
 		AddCommand("rm", "<path> removes object (WARNING: RECURSIVE, be careful!)",
-			make_function([this](const LocalPath &path) -> void { Delete(path); }));
+			make_function([this](const Path &path) -> void { Delete(path); }));
 		AddCommand("rm-id", "<path> removes object by id (WARNING: RECURSIVE, be careful!)",
 			make_function([this](mtp::u32 id) -> void { Delete(mtp::ObjectId(id)); }));
 		AddCommand("mkdir", "<path> makes directory",
@@ -146,7 +159,7 @@ namespace cli
 		AddCommand("mkpath", "<path> create directory structure specified in path",
 			make_function([this](const Path &path) -> void { MakePath(path); }));
 		AddCommand("type", "<path> shows type of file (recognized by libmagic/extension)",
-			make_function([this](const LocalPath &path) -> void { ShowType(path); }));
+			make_function([](const LocalPath &path) -> void { ShowType(path); }));
 
 		AddCommand("rename", "renames object",
 			make_function([this](const Path & path, const std::string & newName) -> void { Rename(path, newName); }));
@@ -161,6 +174,17 @@ namespace cli
 		AddCommand("storage-info", "<storage-id> displays storage information",
 			make_function([this](const StoragePath &path) -> void { DisplayStorageInfo(path); }));
 
+		AddCommand("get-refs", "returns object-associated refs",
+			make_function([this](const StoragePath &path) -> void { GetObjectReferences(path); }));
+
+		if (Library::Supported(_session))
+		{
+			AddCommand("zune-init", "load media library",
+				make_function([this]() -> void { ZuneInit(); }));
+			AddCommand("zune-import", "<file> import file using metadata",
+				make_function([this](const LocalPath &path) -> void { ZuneImport(path); }));
+		}
+
 		AddCommand("test-property-list", "test GetObjectPropList on given object",
 			make_function([this](const Path &path) -> void { TestObjectPropertyList(path); }));
 
@@ -174,6 +198,15 @@ namespace cli
 		}));
 	}
 
+	Session::~Session()
+	{ }
+
+	std::string Session::GetMtpzDataPath()
+	{
+		char * home = getenv("HOME");
+		return std::string(home? home: ".") + "/.mtpz-data";
+	}
+
 	bool Session::SetFirstStorage()
 	{
 		auto ids = _session->GetStorageIDs();
@@ -289,6 +322,11 @@ namespace cli
 		{
 			std::stringstream ss;
 			ss << _gdi.Manufacturer << " " << _gdi.Model;
+			if (_deviceFriendlyNameSupported) {
+				auto name = _session->GetDeviceStringProperty(mtp::DeviceProperty::DeviceFriendlyName);
+				if (!name.empty())
+					ss << " / " << name;
+			}
 			if (_batterySupported)
 			{
 				auto level = _session->GetDeviceIntegerProperty(mtp::DeviceProperty::BatteryLevel);
@@ -315,17 +353,22 @@ namespace cli
 			std::stringstream ss;
 			ss << "supported op codes: ";
 			for(OperationCode code : _gdi.OperationsSupported)
-			{
-				ss << hex(code, 4) << " ";
-			}
-			ss << "\n";
-			ss << "supported properties: ";
-			for(u16 code : _gdi.DevicePropertiesSupported)
-			{
-				ss << hex(code, 4) << " ";
-				if (mtp::DeviceProperty(code) == mtp::DeviceProperty::BatteryLevel)
-					_batterySupported = true;
-			}
+				ss << ToString(code) << " ";
+
+			ss << "\nsupported capture formats: ";
+			for(auto code : _gdi.CaptureFormats)
+				ss << ToString(code) << " ";
+			ss << "\nsupported image formats: ";
+			for(auto code : _gdi.ImageFormats)
+				ss << ToString(code) << " ";
+
+			ss << "\nsupported properties: ";
+			_batterySupported = _gdi.Supports(mtp::DeviceProperty::BatteryLevel);
+			_deviceFriendlyNameSupported = _gdi.Supports(mtp::DeviceProperty::DeviceFriendlyName);
+
+			for(DeviceProperty code : _gdi.DevicePropertiesSupported)
+				ss << ToString(code) << " ";
+
 			ss << "\n";
 			debug(ss.str());
 		}
@@ -492,24 +535,35 @@ namespace cli
 			{
 				try
 				{
-					msg::ObjectInfo info = _session->GetObjectInfo(objectId);
+					std::string filename;
 					if (extended)
+					{
+						msg::ObjectInfo info = _session->GetObjectInfo(objectId);
+						filename = info.Filename;
 						print(
 							std::left,
 							width(objectId, 10), " ",
 							width(info.StorageId.Id, 10), " ",
 							std::right,
-							hex(info.ObjectFormat, 4), " ",
+							ToString(info.ObjectFormat), " ",
 							width(info.ObjectCompressedSize, 10), " ",
 							std::left,
 							width(!info.CaptureDate.empty()? FormatTime(info.CaptureDate): FormatTime(info.ModificationDate), 20), " ",
 							prefix + info.Filename, " "
 						);
+					}
 					else
-						print(std::left, width(objectId, 10), " ", prefix + info.Filename);
+					{
+						filename = _session->GetObjectStringProperty(objectId, ObjectProperty::ObjectFilename);
+						print(std::left, width(objectId, 10), " ", prefix + filename);
+					}
 
-					if (recursive && info.ObjectFormat == mtp::ObjectFormat::Association)
-						List(objectId, extended, recursive, prefix + info.Filename + "/");
+					if (recursive)
+					{
+						auto format = mtp::ObjectFormat(_session->GetObjectIntegerProperty(objectId, ObjectProperty::ObjectFormat));
+						if (format == mtp::ObjectFormat::Association)
+							List(objectId, extended, recursive, prefix + filename + "/");
+					}
 				}
 				catch(const std::exception &ex)
 				{
@@ -772,7 +826,7 @@ namespace cli
 			msg::ObjectInfo oi;
 			oi.Filename = filename;
 			oi.ObjectFormat = ObjectFormatFromFilename(src);
-			oi.SetSize(stream->GetSize());
+			oi.ObjectCompressedSize = stream->GetSize();
 
 			if (_showEvents)
 			{
@@ -805,7 +859,7 @@ namespace cli
 				{
 					auto objectId = ResolveObjectChild(parentDir, filename);
 					ObjectFormat format = ObjectFormat(_session->GetObjectIntegerProperty(objectId, ObjectProperty::ObjectFormat));
-					print("format ", hex((int)format));
+					print("format ", ToString(format));
 					if (format != ObjectFormat::Association)
 						targetFilename = filename; //path exists and it's not directory
 				}
@@ -827,41 +881,89 @@ namespace cli
 
 	mtp::ObjectId Session::MakeDirectory(mtp::ObjectId parentId, const std::string & name)
 	{
-		using namespace mtp;
-		msg::ObjectInfo oi;
-		oi.Filename = name;
-		oi.ObjectFormat = ObjectFormat::Association;
-		auto noi = _session->SendObjectInfo(oi, GetUploadStorageId(), parentId);
+		auto noi = _session->CreateDirectory(name, _cd, GetUploadStorageId());
 		return noi.ObjectId;
 	}
 
 	void Session::ShowType(const LocalPath &src)
 	{
 		mtp::ObjectFormat format = mtp::ObjectFormatFromFilename(src);
-		print("mtp object format = ", hex(format, 4));
+		mtp::print("mtp object format = ", ToString(format));
+	}
+
+	static void PrintFormat(const mtp::ByteArray & format, const mtp::ByteArray & value)
+	{
+		using namespace mtp;
+		InputStream is(format);
+		u16 prop = is.Read16();
+		DataTypeCode type = DataTypeCode(is.Read16());
+		u16 rw = is.Read8();
+
+		std::string defValue;
+		u32 groupCode = 0;
+		u8 formFlag = 0;
+		try
+		{
+			switch(type)
+			{
+#define CASE(BITS) \
+				case mtp::DataTypeCode::Uint##BITS: \
+				case mtp::DataTypeCode::Int##BITS: \
+					defValue = std::to_string(is.Read##BITS ()); break
+				CASE(8); CASE(16); CASE(32); CASE(64); CASE(128);
+#undef CASE
+				case mtp::DataTypeCode::String:
+					defValue = is.ReadString(); break;
+				case mtp::DataTypeCode::ArrayUint8:
+					{
+						u32 size = is.Read32();
+						std::stringstream ss;
+						HexDump(ss, "raw bytes", size, is);
+						defValue = ss.str();
+					}
+					break;
+				default:
+					throw std::runtime_error("invalid type " + std::to_string((u16)type));
+			}
+			groupCode = is.Read32();
+			formFlag = is.Read8();
+		}
+		catch(const std::exception & ex)
+		{ defValue = "<unknown type>"; }
+
+		print("property ", mtp::ToString(ObjectProperty(prop)), ", type: ", ToString(type), ", rw: ", rw, ", default: ", defValue, ", groupCode: ", groupCode, ", form flag: ", formFlag, ", value: ", ToString(type, value));
+		//HexDump("raw", format, true);
 	}
 
 	void Session::ListProperties(mtp::ObjectId id)
 	{
-		auto ops = _session->GetObjectPropertiesSupported(id);
-		std::stringstream ss;
-		ss << "properties supported: ";
+		mtp::ObjectFormat format = mtp::ObjectFormat(_session->GetObjectIntegerProperty(id, mtp::ObjectProperty::ObjectFormat));
+		mtp::print("querying supported properties for format ", mtp::ToString(format));
+
+		auto ops = _session->GetObjectPropertiesSupported(format);
+		mtp::print("properties supported: ");
 		for(mtp::ObjectProperty prop: ops.ObjectPropertyCodes)
 		{
-			ss << mtp::hex(prop, 4) << " ";
+			PrintFormat(_session->GetObjectPropertyDesc(prop), _session->GetObjectProperty(id, prop));
 		}
-		ss << "\n";
-		mtp::print(ss.str());
 	}
 
 	void Session::ListDeviceProperties()
 	{
 		using namespace mtp;
-		for(u16 code : _gdi.DevicePropertiesSupported)
+		for(DeviceProperty code : _gdi.DevicePropertiesSupported)
 		{
-			print("property code: ", hex(code, 4));
-			ByteArray data = _session->GetDeviceProperty((mtp::DeviceProperty)code);
-			HexDump("value", data, true);
+			auto desc = _session->GetDevicePropertyDesc(code);
+			if (code == DeviceProperty::PerceivedDeviceType)
+			{
+				auto value = _session->GetDeviceIntegerProperty(code);
+				print("property: ", ToString(code), " ", ToString(desc.Type), " ", desc.Writeable? "rw ": "ro ", ToString(PerceivedDeviceType(value)));
+			}
+			else
+			{
+				ByteArray value = _session->GetDeviceProperty(code);
+				print("property: ", ToString(code), " ", ToString(desc.Type), " ", desc.Writeable? "rw ": "ro ", ToString(desc.Type, value));
+			}
 		}
 	}
 
@@ -893,7 +995,7 @@ namespace cli
 	void Session::GetObjectPropertyList(mtp::ObjectId parent, const std::set<mtp::ObjectId> &originalObjectList, const mtp::ObjectProperty property)
 	{
 		using namespace mtp;
-		print("testing property 0x", hex(property, 4), "...");
+		print("testing property ", ToString(property), "...");
 
 		std::set<ObjectId> objectList;
 		ByteArray data = _session->GetObjectPropertyList(parent, ObjectFormat::Any, property, 0, 1);
@@ -903,12 +1005,12 @@ namespace cli
 
 		bool ok = true;
 
-		parser.Parse(data, [this, &objectList, property, &ok](mtp::ObjectId objectId, ObjectProperty p, const ByteArray & ) {
+		parser.Parse(data, [&objectList, property, &ok](mtp::ObjectId objectId, ObjectProperty p, const ByteArray & ) {
 			if ((p == property || property == ObjectProperty::All))
 				objectList.insert(objectId);
 			else
 			{
-				print("extra property 0x", hex(p, 4), " returned for object ", objectId.Id, ", while querying property list 0x", mtp::hex(property, 4));
+				print("extra property 0x", ToString(p), " returned for object ", objectId.Id, ", while querying property list ", mtp::ToString(property));
 				ok = false;
 			}
 		});
@@ -918,14 +1020,14 @@ namespace cli
 
 		if (!extraData.empty())
 		{
-			print("inconsistent GetObjectPropertyList for property 0x", mtp::hex(property, 4));
+			print("inconsistent GetObjectPropertyList for property 0x", mtp::ToString(property));
 			for(auto objectId : extraData)
 			{
-				print("missing 0x", mtp::hex(property, 4), " for object ", objectId);
+				print("missing 0x", mtp::ToString(property), " for object ", objectId);
 				ok = false;
 			}
 		}
-		print("getting object property list of type 0x", hex(property, 4), " ", ok? "PASSED": "FAILED");
+		print("getting object property list of type 0x", ToString(property), " ", ok? "PASSED": "FAILED");
 	}
 
 
@@ -968,4 +1070,82 @@ namespace cli
 		print("used ", usedBytes, " (", usedPercents, "%), free ", si.FreeSpaceInBytes, " bytes of ", si.MaxCapacity);
 	}
 
+	void Session::ListObjects(const std::string & format)
+	{ ListObjects(mtp::ObjectFormat(FromHex(format))); }
+
+	void Session::ListObjects(mtp::ObjectFormat format)
+	{
+		mtp::print("querying all objects of format ", mtp::ToString(format));
+		auto objects = _session->GetObjectHandles(mtp::Session::AllStorages, format, mtp::Session::Device);
+		auto & handles = objects.ObjectHandles;
+		for(auto id : handles)
+		{
+			std::string filename, name;
+			try { filename = _session->GetObjectStringProperty(id, mtp::ObjectProperty::ObjectFilename); }
+			catch(const std::exception & ex) { mtp::error("error getting filename: ", ex.what()); }
+			try { name = _session->GetObjectStringProperty(id, mtp::ObjectProperty::Name); }
+			catch(const std::exception & ex) { mtp::error("error getting name: ", ex.what()); }
+			mtp::print(id.Id, "\t", filename, "\t", name);
+		}
+	}
+
+	void Session::GetObjectReferences(const Path & src)
+	{
+		using namespace mtp;
+		auto refs = _session->GetObjectReferences(Resolve(src));
+		for (auto id: refs.ObjectHandles)
+		{
+			debug(id.Id, "\t", _session->GetObjectStringProperty(id, ObjectProperty::ObjectFilename));
+		}
+	}
+
+	void Session::ZuneInit()
+	{
+		if (!_library)
+			_library = std::make_shared<mtp::Library>(_session);
+	}
+
+	void Session::ZuneImport(const LocalPath & path)
+	{
+		ZuneInit();
+		if (!_library)
+			throw std::runtime_error("library failed to initialise");
+
+		using namespace mtp;
+		auto stream = std::make_shared<ObjectInputStream>(path);
+		stream->SetTotal(stream->GetSize());
+		try { stream->SetProgressReporter(ProgressBar(path, _terminalWidth / 3, _terminalWidth)); } catch(const std::exception &ex) { }
+
+		auto meta = Metadata::Read(path);
+		if (!meta)
+			throw std::runtime_error("no metadata");
+
+		print("metadata: ", meta->Artist, " / ", meta->Album, " (", meta->Year, ") / ", meta->Title);
+
+		auto artist = _library->GetArtist(meta->Artist);
+		if (!artist)
+			artist = _library->CreateArtist(meta->Artist);
+		if (!artist)
+			throw std::runtime_error("can't create artist with name " + meta->Artist);
+		debug("got artist record");
+
+		auto album = _library->GetAlbum(artist, meta->Album);
+		if (!album)
+			album = _library->CreateAlbum(artist, meta->Album, meta->Year);
+		if (!album)
+			throw std::runtime_error("can't create album with name " + meta->Album);
+
+		debug("got album record");
+
+		ObjectFormat format = ObjectFormatFromFilename(path);
+		auto slashpos = path.rfind('/');
+		auto filename = slashpos != path.npos? path.substr(slashpos + 1): std::string(path);
+		debug("track format: " + ToString(format));
+		auto songId = _library->CreateTrack(artist, album, format, meta->Title, meta->Genre, meta->Track, filename, stream->GetSize());
+		_session->SendObject(stream);
+
+		_library->AddTrack(album, songId);
+	}
+
+
 }
