@@ -1,6 +1,6 @@
 /*
     This file is part of Android File Transfer For Linux.
-    Copyright (C) 2015-2018  Vladimir Menshakov
+    Copyright (C) 2015-2020  Vladimir Menshakov
 
     This library is free software; you can redistribute it and/or modify it
     under the terms of the GNU Lesser General Public License as published by
@@ -30,10 +30,12 @@
 
 #include <mtp/log.h>
 
+#include <algorithm>
 #include <map>
 #include <set>
 #include <vector>
 #include <functional>
+#include <list>
 #include <stdio.h>
 
 #include <fuse_lowlevel.h>
@@ -45,15 +47,21 @@ namespace
 	class FuseWrapper
 	{
 		std::mutex		_mutex;
+		std::string		_deviceFilter;
 		bool			_claimInterface;
+		bool			_resetDevice;
 		mtp::DevicePtr	_device;
 		mtp::SessionPtr	_session;
 		bool			_editObjectSupported;
 		bool			_getObjectPropertyListSupported;
+		bool			_getPartialObjectSupported;
 		time_t			_connectTime;
+		size_t			_partialObjectCacheSize;
 
-		typedef std::map<std::string, FuseId> ChildrenObjects;
-		typedef std::map<FuseId, ChildrenObjects> Files;
+		static constexpr size_t DefaultPartialObjectCacheSize = 3; //just 3 object entries.
+
+		using ChildrenObjects = std::map<std::string, FuseId>;
+		using Files = std::map<FuseId, ChildrenObjects>;
 		Files			_files;
 
 		typedef std::map<mtp::ObjectId, struct stat> ObjectAttrs;
@@ -73,6 +81,21 @@ namespace
 		std::map<mtp::StorageId, std::string>		_storageToName;
 		std::map<std::string, mtp::StorageId>		_storageFromName;
 
+		struct ObjectCacheEntry
+		{
+			mtp::ObjectId 	Id;
+			mtp::ByteArray 	Data;
+
+			ObjectCacheEntry()
+			{ }
+
+			ObjectCacheEntry(mtp::ObjectId id, mtp::ByteArray && data): Id(id), Data(std::move(data))
+			{ }
+		};
+
+		using ObjectCache = std::list<ObjectCacheEntry>;
+		ObjectCache 								_objectCache;
+
 	private:
 		static FuseId ToFuse(mtp::ObjectId id)
 		{ return FuseId(id.Id + MtpObjectShift); }
@@ -103,7 +126,7 @@ namespace
 
 		void GetObjectInfo(ChildrenObjects &cache, mtp::ObjectId id)
 		{
-			mtp::msg::ObjectInfo oi = _session->GetObjectInfo(id);
+			auto oi = _session->GetObjectInfo(id);
 
 			FuseId inode = ToFuse(id);
 			cache.emplace(oi.Filename, inode);
@@ -280,8 +303,6 @@ namespace
 					catch(const std::exception &ex)
 					{ }
 
-
-
 					return cache;
 				}
 			}
@@ -310,7 +331,7 @@ namespace
 				storageId = _session->GetObjectStorage(parentId);
 			mtp::debug("   creating object in storage ", mtp::hex(storageId.Id), ", parent: ", mtp::hex(parentId.Id, 8));
 
-			mtp::Session::NewObjectInfo noi;
+			mtp::msg::NewObjectInfo noi;
 			if (format != mtp::ObjectFormat::Association)
 			{
 				mtp::msg::ObjectInfo oi;
@@ -377,7 +398,8 @@ namespace
 		}
 
 	public:
-		FuseWrapper(bool claimInterface): _claimInterface(claimInterface)
+		FuseWrapper(const std::string & deviceFilter, bool claimInterface, bool resetDevice):
+			_deviceFilter(deviceFilter), _claimInterface(claimInterface), _resetDevice(resetDevice), _partialObjectCacheSize(DefaultPartialObjectCacheSize)
 		{ Connect(); }
 
 		void Connect()
@@ -390,17 +412,20 @@ namespace
 			_directoryCache.clear();
 			_session.reset();
 			_device.reset();
-			_device = mtp::Device::FindFirst(_claimInterface);
+			_device = mtp::Device::FindFirst(_deviceFilter, _claimInterface, _resetDevice);
 			if (!_device)
 				throw std::runtime_error("no MTP device found");
 
 			_session = _device->OpenSession(1);
 			_editObjectSupported = _session->EditObjectSupported();
 			if (!_editObjectSupported)
-				mtp::error("your device does not have android EditObject extension, mounting read-only\n");
+				mtp::error("your device does not have android EditObject extension, mounting read-only");
 			_getObjectPropertyListSupported = _session->GetObjectPropertyListSupported();
 			if (!_getObjectPropertyListSupported)
-				mtp::error("your device does not have GetObjectPropertyList extension, expect slow enumeration of big directories\n");
+				mtp::error("your device does not have GetObjectPropertyList extension, expect slow enumeration of big directories");
+			_getPartialObjectSupported = _session->GetDeviceInfo().Supports(mtp::OperationCode::GetPartialObject);
+			if (!_getPartialObjectSupported)
+				mtp::error("your device does not have GetPartialObject extension (beware, this will download the latest N objects you accessed and keep them in a small in-memory cache)");
 
 			_connectTime = time(NULL);
 			PopulateStorages();
@@ -430,14 +455,23 @@ namespace
 				_storageToName[id] = path;
 			}
 		}
+		std::string GetMountpoint()
+		{ return _session->GetDeviceInfo().GetFilesystemFriendlyName(); }
 
 		void Init(void *, fuse_conn_info *conn)
 		{
+#if 0
+			/*
+				Disable FUSE_CAP_BIG_WRITES until further investigation
+				https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/commit/?h=v5.4&id=1fb027d7596464d3fad3ed59f70f43807ef926c6
+				https://github.com/netheril96/securefs/commit/e5b9e5d41a6dcdee60ab4a18b46b9a0c6475ba21
+			*/
 			mtp::scoped_mutex_lock l(_mutex);
 			conn->want |= conn->capable & FUSE_CAP_BIG_WRITES; //big writes
 			static const size_t MaxWriteSize = 1024 * 1024;
 			if (conn->max_write < MaxWriteSize)
 				conn->max_write = MaxWriteSize;
+#endif
 		}
 
 		void Lookup (fuse_req_t req, FuseId parent, const char *name)
@@ -521,9 +555,48 @@ namespace
 			off_t rsize = std::min<off_t>(attr.st_size - begin, size);
 			mtp::debug("reading ", rsize, " bytes");
 			mtp::ByteArray data;
+			auto objectId = FromFuse(ino);
 			if (rsize > 0)
-				data = _session->GetPartialObject(FromFuse(ino), begin, rsize);
-			mtp::debug("read", data.size(), "bytes of data");
+			{
+				if (!_getPartialObjectSupported)
+				{
+					std::size_t size = 0;
+
+					ObjectCache::iterator it;
+					for(it = _objectCache.begin(); it != _objectCache.end(); ++it, ++size)
+					{
+						auto & entry = *it;
+						if (entry.Id == objectId)
+						{
+							mtp::debug("in-memory cache hit");
+							auto src = entry.Data.data() + begin;
+							std::copy(src, src + rsize, std::back_inserter(data));
+							if (it != _objectCache.begin())
+								std::swap(*_objectCache.begin(), *it);
+							break;
+						}
+					}
+					if (it == _objectCache.end())
+					{
+						mtp::debug("in-memory cache miss");
+						auto stream = std::make_shared<mtp::ByteArrayObjectOutputStream>();
+						_session->GetObject(objectId, stream);
+						auto src = stream->GetData().data() + begin;
+						std::copy(src, src + rsize, std::back_inserter(data));
+						_objectCache.emplace_front(objectId, std::move(stream->GetData()));
+						++size;
+					}
+
+					while (size > _partialObjectCacheSize)
+					{
+						mtp::debug("purging last entry from cache...");
+						_objectCache.pop_back();
+						--size;
+					}
+				} else
+					data = _session->GetPartialObject(objectId, begin, rsize);
+			}
+			mtp::debug("read ", data.size(), " bytes of data");
 			FUSE_CALL(fuse_reply_buf(req, static_cast<char *>(static_cast<void *>(data.data())), data.size()));
 		}
 
@@ -591,11 +664,6 @@ namespace
 			FUSE_CALL(fuse_reply_err(req, 0));
 		}
 
-		void Rename(fuse_req_t req, FuseId parent, const char *name, FuseId newparent, const char *newname)
-		{
-			fuse_reply_err(req, EXDEV); //return cross-device link, so user space should re-create file and copy it
-		}
-
 		void SetAttr(fuse_req_t req, FuseId inode, struct stat *attr, int to_set, struct fuse_file_info *fi)
 		{
 			mtp::scoped_mutex_lock l(_mutex);
@@ -616,8 +684,14 @@ namespace
 				entry.ReplyError(ENOENT);
 		}
 
-		void RemoveDir (fuse_req_t req, FuseId parent, const char *name)
-		{ Unlink(req, parent, name); }
+		void UnlinkImpl(FuseId inode)
+		{
+			mtp::debug("   unlinking inode ", inode.Inode);
+			mtp::ObjectId id = FromFuse(inode);
+			_session->DeleteObject(id);
+			_openedFiles.erase(inode);
+			_objectAttrs.erase(id);
+		}
 
 		void Unlink(fuse_req_t req, FuseId parent, const char *name)
 		{
@@ -631,14 +705,52 @@ namespace
 			}
 
 			FuseId inode = i->second;
-			mtp::debug("   unlinking inode ", inode.Inode);
-			mtp::ObjectId id = FromFuse(inode);
+			UnlinkImpl(inode);
 			_directoryCache.erase(parent);
-			_openedFiles.erase(inode);
-			_objectAttrs.erase(id);
 			children.erase(i);
 
-			_session->DeleteObject(id);
+			FUSE_CALL(fuse_reply_err(req, 0));
+		}
+
+		void RemoveDir (fuse_req_t req, FuseId parent, const char *name)
+		{ Unlink(req, parent, name); }
+
+		void Rename(fuse_req_t req, FuseId parent, const char *name, FuseId newparent, const char *newName)
+		{
+			if (parent != newparent) {
+				//no renames across directory boundary, sorry
+				//return cross-device link, so user space should re-create file and copy it
+				FUSE_CALL(fuse_reply_err(req, EXDEV));
+				return;
+			}
+
+			mtp::scoped_mutex_lock l(_mutex);
+			ChildrenObjects &children = GetChildren(parent);
+			auto i = children.find(name);
+			if (i == children.end())
+			{
+				FUSE_CALL(fuse_reply_err(req, ENOENT));
+				return;
+			}
+
+			FuseId inode = i->second;
+			mtp::debug("   renaming inode ", inode.Inode, " to ", newName);
+
+			auto old = children.find(newName);
+			if (old != children.end())
+			{
+				mtp::debug("   unlinking target inode ", old->second.Inode);
+				UnlinkImpl(old->second);
+				children.erase(old);
+			}
+
+			mtp::ObjectId id = FromFuse(inode);
+			_session->SetObjectProperty(id, mtp::ObjectProperty::ObjectFilename, std::string(newName));
+
+			children.erase(i);
+			children.emplace(newName, inode);
+			_directoryCache.erase(parent);
+
 			FUSE_CALL(fuse_reply_err(req, 0));
 		}
 
@@ -758,19 +870,47 @@ namespace
 
 int main(int argc, char **argv)
 {
+	std::string deviceFilter;
 	bool claimInterface = true;
+	bool resetDevice = false;
+	bool showHelp = false;
+
+	std::vector<char *> args;
+	args.push_back(argv[0]);
+
 	for(int i = 1; i < argc; ++i)
 	{
-		if (strcmp(argv[i], "-C") == 0)
+		if (strcmp(argv[i], "-R") == 0)
+			resetDevice = false;
+		else if (strcmp(argv[i], "-C") == 0)
 			claimInterface = false;
-		if (strcmp(argv[i], "-d") == 0 || strcmp(argv[i], "-odebug") == 0)
+		else if (strcmp(argv[i], "-d") == 0 || strcmp(argv[i], "-odebug") == 0)
+		{
+			args.push_back(argv[i]);
 			mtp::g_debug = true;
-		if (i + 1 < argc && strcmp(argv[i], "-o") == 0 && strcmp(argv[i + 1], "debug") == 0)
+		}
+		else if (i + 1 < argc && strcmp(argv[i], "-o") == 0 && strcmp(argv[i + 1], "debug") == 0)
+		{
 			mtp::g_debug = true;
+			args.push_back(argv[i]);
+		}
+		else if (i + 1 < argc && strcmp(argv[i], "-D") == 0)
+		{
+			deviceFilter = argv[i + 1];
+			++i;
+		}
+		else if (strcmp(argv[i], "--help") == 0 || strcmp(argv[i], "-h") == 0)
+		{
+			args.push_back(argv[i]);
+			showHelp = true;
+		} else
+			args.push_back(argv[i]);
 	}
 
+	args.push_back(nullptr);
+
 	try
-	{ g_wrapper.reset(new FuseWrapper(claimInterface)); }
+	{ g_wrapper.reset(new FuseWrapper(deviceFilter, claimInterface, resetDevice)); }
 	catch(const std::exception &ex)
 	{ mtp::error("connect failed: ", ex.what()); return 1; }
 
@@ -793,34 +933,50 @@ int main(int argc, char **argv)
 	ops.unlink		= &Unlink;
 	ops.statfs		= &StatFS;
 
-	struct fuse_args args = FUSE_ARGS_INIT(argc, argv);
+	struct fuse_args fuse_args = FUSE_ARGS_INIT(static_cast<int>(args.size() - 1), args.data());
 	struct fuse_chan *ch;
 	char *mountpoint;
 	int err = -1;
 	int multithreaded = 0, foreground = 0;
 
-	if (fuse_parse_cmdline(&args, &mountpoint, &multithreaded, &foreground) != -1 &&
-	    mountpoint != NULL &&
-	    (ch = fuse_mount(mountpoint, &args)) != NULL) {
-		struct fuse_session *se;
-
-		se = fuse_lowlevel_new(&args, &ops,
-				       sizeof(ops), NULL);
-		if (se != NULL) {
-			if (fuse_set_signal_handlers(se) != -1)
-			{
-				fuse_session_add_chan(se, ch);
-				if (fuse_daemonize(foreground) == -1)
-					perror("fuse_daemonize");
-				err = (multithreaded? fuse_session_loop_mt: fuse_session_loop)(se);
-				fuse_remove_signal_handlers(se);
-				fuse_session_remove_chan(ch);
-			}
-			fuse_session_destroy(se);
+	if (fuse_parse_cmdline(&fuse_args, &mountpoint, &multithreaded, &foreground) != -1)
+	{
+		if (!mountpoint)
+		{
+			auto mp = g_wrapper->GetMountpoint();
+			mountpoint = strdup(mp.c_str());
+			mkdir(mountpoint, 0700);
 		}
-		fuse_unmount(mountpoint, ch);
+
+	    if (mountpoint != NULL && (ch = fuse_mount(mountpoint, &fuse_args)) != NULL)
+		{
+			struct fuse_session *se = fuse_lowlevel_new(&fuse_args, &ops, sizeof(ops), NULL);
+			if (se != NULL)
+			{
+				if (fuse_set_signal_handlers(se) != -1)
+				{
+					fuse_session_add_chan(se, ch);
+					if (fuse_daemonize(foreground) == -1)
+						perror("fuse_daemonize");
+					err = (multithreaded? fuse_session_loop_mt: fuse_session_loop)(se);
+					fuse_remove_signal_handlers(se);
+					fuse_session_remove_chan(ch);
+				}
+				fuse_session_destroy(se);
+			}
+			fuse_unmount(mountpoint, ch);
+		}
+	} else {
+		mtp::error("fuse_parse_cmdline failed");
 	}
-	fuse_opt_free_args(&args);
+	if (showHelp) {
+		mtp::print();
+		mtp::print("Additional AFT options: ");
+		mtp::print("    -R                     reset device");
+		mtp::print("    -C                     do not claim USB interface");
+		mtp::print("    -v                     AFT verbose output");
+	}
+	fuse_opt_free_args(&fuse_args);
 
 	return err ? 1 : 0;
 }
