@@ -1,6 +1,6 @@
 /*
     This file is part of Android File Transfer For Linux.
-    Copyright (C) 2015-2018  Vladimir Menshakov
+    Copyright (C) 2015-2020  Vladimir Menshakov
 
     This library is free software; you can redistribute it and/or modify it
     under the terms of the GNU Lesser General Public License as published by
@@ -17,8 +17,8 @@
     Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
 
-#ifndef BYTEARRAY_H
-#define	BYTEARRAY_H
+#ifndef AFTL_MTP_BYTEARRAY_H
+#define AFTL_MTP_BYTEARRAY_H
 
 #include <mtp/types.h>
 #include <vector>
@@ -28,9 +28,6 @@ namespace mtp
 {
 	using ByteArray = std::vector<u8>; //! \typedef ByteArray : vector of bytes
 	DECLARE_PTR(ByteArray);
-
-	//! output hex dump to debug channel
-	void HexDump(const std::string &prefix, const ByteArray &data, bool force = false);
 }
 
 #endif	/* BYTEARRAY_H */
