@@ -1,6 +1,6 @@
 /*
     This file is part of Android File Transfer For Linux.
-    Copyright (C) 2015-2018  Vladimir Menshakov
+    Copyright (C) 2015-2020  Vladimir Menshakov
 
     This library is free software; you can redistribute it and/or modify it
     under the terms of the GNU Lesser General Public License as published by
@@ -42,14 +42,18 @@ namespace mtp
 		throw InvalidResponseException(__func__, (RCODE)); \
 } while(false)
 
-	Session::Session(usb::BulkPipePtr pipe, u32 sessionId):
-		_packeter(pipe), _sessionId(sessionId), _nextTransactionId(1), _transaction(),
-		_getObjectModificationTimeBuggy(false),
+	Session::Session(const PipePacketer & packeter, u32 sessionId):
+		_packeter(packeter), _sessionId(sessionId), _nextTransactionId(1), _transaction(),
+		_getObjectModificationTimeBuggy(false), _separateBulkWrites(false),
 		_defaultTimeout(DefaultTimeout)
 	{
-		_deviceInfo = GetDeviceInfoImpl();
+		_deviceInfo = GetDeviceInfo(_packeter, _defaultTimeout);
+		if (_deviceInfo.Manufacturer == "Microsoft")
+			_separateBulkWrites = true;
+
 		_getPartialObject64Supported = _deviceInfo.Supports(OperationCode::GetPartialObject64);
 		_getObjectPropertyListSupported = _deviceInfo.Supports(OperationCode::GetObjectPropList);
+		_getObjectPropValueSupported = _deviceInfo.Supports(OperationCode::GetObjectPropValue);
 		_editObjectSupported = _deviceInfo.Supports(OperationCode::BeginEditObject) &&
 			_deviceInfo.Supports(OperationCode::EndEditObject) &&
 			_deviceInfo.Supports(OperationCode::TruncateObject) &&
@@ -79,13 +83,20 @@ namespace mtp
 			_transaction->Id = _nextTransactionId++;
 	}
 
+	void Session::Send(PipePacketer &packeter, const OperationRequest &req, int timeout)
+	{
+		if (timeout <= 0)
+			timeout = DefaultTimeout;
+		Container container(req);
+		packeter.Write(container.Data, timeout);
+	}
+
 
 	void Session::Send(const OperationRequest &req, int timeout)
 	{
 		if (timeout <= 0)
 			timeout = _defaultTimeout;
-		Container container(req);
-		_packeter.Write(container.Data, timeout);
+		Send(_packeter, req, timeout);
 	}
 
 	void Session::Close()
@@ -98,102 +109,122 @@ namespace mtp
 		//HexDump("payload", data);
 	}
 
-	ByteArray Session::Get(u32 transaction, int timeout)
+	ByteArray Session::Get(PipePacketer &packeter, u32 transaction, ByteArray & response, int timeout)
 	{
 		if (timeout <= 0)
-			timeout = _defaultTimeout;
-		ByteArray data, response;
+			timeout = DefaultTimeout;
+
+		ByteArray data;
 		ResponseType responseCode;
-		_packeter.Read(transaction, data, responseCode, response, timeout);
+		packeter.Read(transaction, data, responseCode, response, timeout);
 		CHECK_RESPONSE(responseCode);
 		return data;
 	}
 
-	template<typename ... Args>
-	ByteArray Session::RunTransactionWithDataRequest(int timeout, OperationCode code, const IObjectInputStreamPtr & inputStream, Args && ... args)
+	ByteArray Session::Get(u32 transaction, ByteArray &response, int timeout)
 	{
+		if (timeout <= 0)
+			timeout = _defaultTimeout;
+		return Get(_packeter, transaction, response, timeout);
+	}
+
+	template<typename ... Args>
+	ByteArray Session::RunTransactionWithDataRequest(int timeout, OperationCode code, ByteArray & response, const IObjectInputStreamPtr & inputStream, Args && ... args)
+	{
+#if 0
+		try
+		{ _packeter.PollEvent(_defaultTimeout); }
+		catch(const std::exception &ex)
+		{ error("exception in interrupt: ", ex.what()); }
+#endif
+
 		scoped_mutex_lock l(_mutex);
+		if (!_deviceInfo.Supports(code))
+			throw std::runtime_error("Operation code " + ToString(code) + " not supported.");
 		Transaction transaction(this);
 		Send(OperationRequest(code, transaction.Id, std::forward<Args>(args) ... ), timeout);
 		if (inputStream)
 		{
 			DataRequest req(code, transaction.Id);
 			Container container(req, inputStream);
-			_packeter.Write(std::make_shared<JoinedObjectInputStream>(std::make_shared<ByteArrayObjectInputStream>(container.Data), inputStream), timeout);
+			if (_separateBulkWrites)
+			{
+				_packeter.Write(std::make_shared<ByteArrayObjectInputStream>(container.Data), timeout);
+				_packeter.Write(inputStream, timeout);
+			} else
+				_packeter.Write(std::make_shared<JoinedObjectInputStream>(std::make_shared<ByteArrayObjectInputStream>(container.Data), inputStream), timeout);
 		}
-		return Get(transaction.Id);
+		return Get(transaction.Id, response);
 	}
 
-	msg::DeviceInfo Session::GetDeviceInfoImpl()
+	msg::DeviceInfo Session::GetDeviceInfo(PipePacketer& packeter, int timeout)
 	{
-		auto data = RunTransaction(_defaultTimeout, OperationCode::GetDeviceInfo);
-		InputStream stream(data); //operation code + session id
-		msg::DeviceInfo gdi;
-		gdi.Read(stream);
-		return gdi;
+		ByteArray response;
+		Send(packeter, OperationRequest(OperationCode::GetDeviceInfo, 0), timeout);
+		auto info = Get(packeter, 0, response, timeout);
+		return ParseResponse<msg::DeviceInfo>(info);
 	}
 
 
 	msg::ObjectHandles Session::GetObjectHandles(StorageId storageId, ObjectFormat objectFormat, ObjectId parent, int timeout)
-	{
-		auto data = RunTransaction(timeout, OperationCode::GetObjectHandles, storageId.Id, static_cast<u32>(objectFormat), parent.Id);
-		InputStream stream(data);
-
-		msg::ObjectHandles goh;
-		goh.Read(stream);
-		return goh;
-	}
+	{ return ParseResponse<msg::ObjectHandles>(RunTransaction(timeout, OperationCode::GetObjectHandles, storageId.Id, static_cast<u32>(objectFormat), parent.Id)); }
 
 	msg::StorageIDs Session::GetStorageIDs()
-	{
-		auto data = RunTransaction(_defaultTimeout, OperationCode::GetStorageIDs);
-		InputStream stream(data);
-
-		msg::StorageIDs gsi;
-		gsi.Read(stream);
-		return gsi;
-	}
+	{ return ParseResponse<msg::StorageIDs>(RunTransaction(_defaultTimeout, OperationCode::GetStorageIDs)); }
 
 	msg::StorageInfo Session::GetStorageInfo(StorageId storageId)
+	{ return ParseResponse<msg::StorageInfo>(RunTransaction(_defaultTimeout, OperationCode::GetStorageInfo, storageId.Id)); }
+
+	msg::SendObjectPropListResponse Session::SendObjectPropList(StorageId storageId, ObjectId parentId, ObjectFormat format, u64 objectSize, const ByteArray & propList)
 	{
-		auto data = RunTransaction(_defaultTimeout, OperationCode::GetStorageInfo, storageId.Id);
-		InputStream stream(data);
-		msg::StorageInfo gsi;
-		gsi.Read(stream);
-		return gsi;
+		ByteArray responseData;
+		IObjectInputStreamPtr inputStream = std::make_shared<ByteArrayObjectInputStream>(propList);
+		RunTransactionWithDataRequest(_defaultTimeout, OperationCode::SendObjectPropList, responseData, inputStream, storageId.Id, parentId.Id, static_cast<u32>(format), static_cast<u32>(objectSize >> 32), static_cast<u32>(objectSize));
+		return ParseResponse<msg::SendObjectPropListResponse>(responseData);
 	}
 
-	Session::NewObjectInfo Session::CreateDirectory(const std::string &name, ObjectId parentId, StorageId storageId, AssociationType type)
+	msg::NewObjectInfo Session::CreateDirectory(const std::string &name, ObjectId parentId, StorageId storageId, AssociationType type)
 	{
-		mtp::msg::ObjectInfo oi;
-		oi.Filename = name;
-		oi.ParentObject = parentId;
-		oi.StorageId = storageId;
-		oi.ObjectFormat = mtp::ObjectFormat::Association;
-		oi.AssociationType = type;
-		mtp::Session::NewObjectInfo noi = SendObjectInfo(oi, storageId, parentId);
+		if (_deviceInfo.Supports(OperationCode::SendObjectPropList))
+		{
+			//modern way of creating objects
+			ByteArray propList;
+			{
+				OutputStream os(propList);
+				os.Write32(1); //number of props
+				os.Write32(0); //object handle
+				os.Write16(static_cast<u16>(ObjectProperty::ObjectFilename));
+				os.Write16(static_cast<u16>(DataTypeCode::String));
+				os.WriteString(name);
+			}
+			auto response = SendObjectPropList(storageId, parentId, ObjectFormat::Association, 0, propList);
 
-		//SendObject(std::make_shared<ByteArrayObjectInputStream>(ByteArray()));
-		return noi;
+			msg::NewObjectInfo noi;
+			noi.StorageId = response.StorageId;
+			noi.ParentObjectId = response.ParentObjectId;
+			noi.ObjectId = response.ObjectId;
+			return noi;
+		}
+		else
+		{
+			mtp::msg::ObjectInfo oi;
+			oi.Filename = name;
+			oi.ParentObject = parentId;
+			oi.StorageId = storageId;
+			oi.ObjectFormat = mtp::ObjectFormat::Association;
+			oi.AssociationType = type;
+			return SendObjectInfo(oi, storageId, parentId);
+		}
 	}
 
 	msg::ObjectInfo Session::GetObjectInfo(ObjectId objectId)
-	{
-		auto data = RunTransaction(_defaultTimeout, OperationCode::GetObjectInfo, objectId.Id);
-		InputStream stream(data);
-		msg::ObjectInfo goi;
-		goi.Read(stream);
-		return goi;
-	}
+	{ return ParseResponse<msg::ObjectInfo>(RunTransaction(_defaultTimeout, OperationCode::GetObjectInfo, objectId.Id)); }
 
-	msg::ObjectPropertiesSupported Session::GetObjectPropertiesSupported(ObjectId objectId)
-	{
-		auto data = RunTransaction(_defaultTimeout, OperationCode::GetObjectPropsSupported, objectId.Id);
-		InputStream stream(data);
-		msg::ObjectPropertiesSupported ops;
-		ops.Read(stream);
-		return ops;
-	}
+	msg::ObjectPropertiesSupported Session::GetObjectPropertiesSupported(ObjectFormat format)
+	{ return ParseResponse<msg::ObjectPropertiesSupported>(RunTransaction(_defaultTimeout, OperationCode::GetObjectPropsSupported, static_cast<u32>(format))); }
+
+	ByteArray Session::GetObjectPropertyDesc(ObjectProperty code)
+	{ return RunTransaction(_defaultTimeout, OperationCode::GetObjectPropDesc, static_cast<u32>(code)); }
 
 	void Session::GetObject(ObjectId objectId, const IObjectOutputStreamPtr &outputStream)
 	{
@@ -230,10 +261,33 @@ namespace mtp
 	}
 
 
-	Session::NewObjectInfo Session::SendObjectInfo(const msg::ObjectInfo &objectInfo, StorageId storageId, ObjectId parentObject)
+	msg::NewObjectInfo Session::SendObjectInfo(const msg::ObjectInfo &objectInfo, StorageId storageId, ObjectId parentObject)
 	{
 		if (objectInfo.Filename.empty())
 			throw std::runtime_error("object filename must not be empty");
+
+		if (_deviceInfo.Supports(OperationCode::SendObjectPropList))
+		{
+			//modern way of creating objects
+			ByteArray propList;
+			{
+				OutputStream os(propList);
+				os.Write32(1); //number of props
+				os.Write32(0); //object handle
+				os.Write16(static_cast<u16>(ObjectProperty::ObjectFilename));
+				os.Write16(static_cast<u16>(DataTypeCode::String));
+				os.WriteString(objectInfo.Filename);
+			}
+
+			auto response = SendObjectPropList(storageId, parentObject, objectInfo.ObjectFormat, objectInfo.ObjectCompressedSize, propList);
+
+			msg::NewObjectInfo noi;
+			noi.StorageId = response.StorageId;
+			noi.ParentObjectId = response.ParentObjectId;
+			noi.ObjectId = response.ObjectId;
+			return noi;
+		}
+
 		scoped_mutex_lock l(_mutex);
 		Transaction transaction(this);
 		Send(OperationRequest(OperationCode::SendObjectInfo, transaction.Id, storageId.Id, parentObject.Id));
@@ -249,12 +303,7 @@ namespace mtp
 		_packeter.Read(transaction.Id, data, responseCode, response, _defaultTimeout);
 		//HexDump("response", response);
 		CHECK_RESPONSE(responseCode);
-		InputStream stream(response);
-		NewObjectInfo noi;
-		stream >> noi.StorageId;
-		stream >> noi.ParentObjectId;
-		stream >> noi.ObjectId;
-		return noi;
+		return ParseResponse<msg::NewObjectInfo>(response);
 	}
 
 	void Session::SendObject(const IObjectInputStreamPtr &inputStream, int timeout)
@@ -265,9 +314,15 @@ namespace mtp
 		{
 			DataRequest req(OperationCode::SendObject, transaction.Id);
 			Container container(req, inputStream);
-			_packeter.Write(std::make_shared<JoinedObjectInputStream>(std::make_shared<ByteArrayObjectInputStream>(container.Data), inputStream), timeout);
+			if (_separateBulkWrites)
+			{
+				_packeter.Write(std::make_shared<ByteArrayObjectInputStream>(container.Data), timeout);
+				_packeter.Write(inputStream, timeout);
+			} else
+				_packeter.Write(std::make_shared<JoinedObjectInputStream>(std::make_shared<ByteArrayObjectInputStream>(container.Data), inputStream), timeout);
 		}
-		Get(transaction.Id);
+		ByteArray response;
+		Get(transaction.Id, response);
 	}
 
 	void Session::BeginEditObject(ObjectId objectId)
@@ -276,7 +331,8 @@ namespace mtp
 	void Session::SendPartialObject(ObjectId objectId, u64 offset, const ByteArray &data)
 	{
 		IObjectInputStreamPtr inputStream = std::make_shared<ByteArrayObjectInputStream>(data);
-		RunTransactionWithDataRequest(_defaultTimeout, OperationCode::SendPartialObject, inputStream, objectId.Id, offset, offset >> 32, data.size());
+		ByteArray response;
+		RunTransactionWithDataRequest(_defaultTimeout, OperationCode::SendPartialObject, response, inputStream, objectId.Id, offset, offset >> 32, data.size());
 	}
 
 	void Session::TruncateObject(ObjectId objectId, u64 size)
@@ -288,7 +344,8 @@ namespace mtp
 	void Session::SetObjectProperty(ObjectId objectId, ObjectProperty property, const ByteArray &value)
 	{
 		IObjectInputStreamPtr inputStream = std::make_shared<ByteArrayObjectInputStream>(value);
-		RunTransactionWithDataRequest(_defaultTimeout, OperationCode::SetObjectPropValue, inputStream, objectId.Id, (u16)property);
+		ByteArray response;
+		RunTransactionWithDataRequest(_defaultTimeout, OperationCode::SetObjectPropValue, response, inputStream, objectId.Id, (u16)property);
 	}
 
 	StorageId Session::GetObjectStorage(mtp::ObjectId id)
@@ -315,7 +372,45 @@ namespace mtp
 	{ return RunTransaction(_defaultTimeout, OperationCode::GetObjectPropValue, objectId.Id, (u16)property); }
 
 	u64 Session::GetObjectIntegerProperty(ObjectId objectId, ObjectProperty property)
-	{ return ReadSingleInteger(GetObjectProperty(objectId, property)); }
+	{
+		if (!_getObjectPropValueSupported) {
+			auto info = GetObjectInfo(objectId);
+			switch(property)
+			{
+				case ObjectProperty::StorageId:
+					return info.StorageId.Id;
+				case ObjectProperty::ObjectFormat:
+					return static_cast<u32>(info.ObjectFormat);
+				case ObjectProperty::ProtectionStatus:
+					return info.ProtectionStatus;
+				case ObjectProperty::ObjectSize:
+					return info.ObjectCompressedSize;
+				case ObjectProperty::RepresentativeSampleFormat:
+					return info.ThumbFormat;
+				case ObjectProperty::RepresentativeSampleSize:
+					return info.ThumbCompressedSize;
+				case ObjectProperty::RepresentativeSampleWidth:
+					return info.ThumbPixWidth;
+				case ObjectProperty::RepresentativeSampleHeight:
+					return info.ThumbPixHeight;
+				case ObjectProperty::Width:
+					return info.ImagePixWidth;
+				case ObjectProperty::Height:
+					return info.ImagePixHeight;
+				case ObjectProperty::ImageBitDepth:
+					return info.ImageBitDepth;
+				case ObjectProperty::ParentObject:
+					return info.ParentObject.Id;
+				case ObjectProperty::AssociationType:
+					return static_cast<u32>(info.AssociationType);
+				case ObjectProperty::AssociationDesc:
+					return info.AssociationDesc;
+				default:
+					throw std::runtime_error("Device does not support object properties and no ObjectInfo fallback for " + ToString(property) + ".");
+			}
+		}
+		return ReadSingleInteger(GetObjectProperty(objectId, property));
+	}
 
 	void Session::SetObjectProperty(ObjectId objectId, ObjectProperty property, u64 value)
 	{
@@ -334,8 +429,36 @@ namespace mtp
 		SetObjectProperty(objectId, property, ByteArray(data.begin(), data.begin() + i));
 	}
 
+	void Session::SetObjectPropertyAsArray(ObjectId objectId, ObjectProperty property, const ByteArray &value)
+	{
+		auto n = value.size();
+
+		ByteArray array;
+		OutputStream out(array);
+		array.reserve(n + 4);
+		out.WriteArray(value);
+
+		SetObjectProperty(objectId, property, array);
+	}
+
 	std::string Session::GetObjectStringProperty(ObjectId objectId, ObjectProperty property)
 	{
+		if (!_getObjectPropValueSupported) {
+			auto info = GetObjectInfo(objectId);
+			switch(property)
+			{
+				case ObjectProperty::ObjectFilename:
+					return info.Filename;
+				case ObjectProperty::DateCreated:
+				case ObjectProperty::DateAuthored:
+				case ObjectProperty::DateAdded:
+					return info.CaptureDate;
+				case ObjectProperty::DateModified:
+					return info.ModificationDate;
+				default:
+					throw std::runtime_error("Device does not support object properties and no ObjectInfo fallback for " + ToString(property) + ".");
+			}
+		}
 		return ReadSingleString(GetObjectProperty(objectId, property));
 	}
 
@@ -366,18 +489,47 @@ namespace mtp
 	std::string Session::GetDeviceStringProperty(DeviceProperty property)
 	{ return ReadSingleString(GetDeviceProperty(property)); }
 
-	ByteArray Session::GetObjectPropertyList(ObjectId objectId, ObjectFormat format, ObjectProperty property, u32 groupCode, u32 depth, int timeout)
+	void Session::SetDeviceProperty(DeviceProperty property, const ByteArray &value)
 	{
-		if (objectId == Root) //ffffffff -> 0
-			objectId = Device;
-		return RunTransaction(timeout, OperationCode::GetObjectPropList, objectId.Id, (u32)format, property != ObjectProperty::All? (u32)property: 0xffffffffu, groupCode, depth);
+		IObjectInputStreamPtr inputStream = std::make_shared<ByteArrayObjectInputStream>(value);
+		ByteArray response;
+		RunTransactionWithDataRequest(_defaultTimeout, OperationCode::SetDevicePropValue, response, inputStream, (u16)property);
 	}
 
+	void Session::SetDeviceProperty(DeviceProperty property, const std::string &value)
+	{
+		ByteArray data;
+		data.reserve(value.size() * 2 + 1);
+		OutputStream stream(data);
+		stream << value;
+		SetDeviceProperty(property, data);
+	}
+
+
+	ByteArray Session::GetObjectPropertyList(ObjectId objectId, ObjectFormat format, ObjectProperty property, u32 groupCode, u32 depth, int timeout)
+	{ return RunTransaction(timeout, OperationCode::GetObjectPropList, objectId.Id, (u32)format, property != ObjectProperty::All? (u32)property: 0xffffffffu, groupCode, depth); }
+
 	void Session::DeleteObject(ObjectId objectId, int timeout)
 	{ RunTransaction(timeout, OperationCode::DeleteObject, objectId.Id, 0); }
 
+	msg::DevicePropertyDesc Session::GetDevicePropertyDesc(DeviceProperty property)
+	{ return ParseResponse<msg::DevicePropertyDesc>(RunTransaction(_defaultTimeout, OperationCode::GetDevicePropDesc, static_cast<u32>(property))); }
+
 	ByteArray Session::GetDeviceProperty(DeviceProperty property)
-	{ return RunTransaction(_defaultTimeout, OperationCode::GetDevicePropValue, (u16)property); }
+	{ return RunTransaction(_defaultTimeout, OperationCode::GetDevicePropValue, static_cast<u32>(property)); }
+
+	ByteArray Session::GenericOperation(OperationCode code)
+	{ return RunTransaction(_defaultTimeout, code); }
+
+	ByteArray Session::GenericOperation(OperationCode code, const ByteArray & payload)
+	{
+		IObjectInputStreamPtr inputStream = std::make_shared<ByteArrayObjectInputStream>(payload);
+		ByteArray response;
+		return RunTransactionWithDataRequest(_defaultTimeout, code, response, inputStream);
+	}
+
+	void Session::EnableSecureFileOperations(u32 cmac[4])
+	{ RunTransaction(_defaultTimeout, OperationCode::EnableTrustedFilesOperations, cmac[0], cmac[1], cmac[2], cmac[3]); }
 
 	Session::ObjectEditSession::ObjectEditSession(const SessionPtr & session, ObjectId objectId): _session(session), _objectId(objectId)
 	{ session->BeginEditObject(objectId); }
@@ -403,4 +555,18 @@ namespace mtp
 		_packeter.Abort(transactionId, timeout);
 	}
 
+	void Session::SetObjectReferences(ObjectId objectId, const msg::ObjectHandles & objects)
+	{
+		ByteArray data;
+		OutputStream out(data);
+		objects.Write(out);
+
+		IObjectInputStreamPtr inputStream = std::make_shared<ByteArrayObjectInputStream>(data);
+		ByteArray response;
+		RunTransactionWithDataRequest(_defaultTimeout, OperationCode::SetObjectReferences, response, inputStream, objectId.Id);
+	}
+
+	msg::ObjectHandles Session::GetObjectReferences(ObjectId objectId)
+	{ return ParseResponse<msg::ObjectHandles>(RunTransaction(_defaultTimeout, OperationCode::GetObjectReferences, objectId.Id)); }
+
 }
