@@ -1,6 +1,6 @@
 /*
     This file is part of Android File Transfer For Linux.
-    Copyright (C) 2015-2018  Vladimir Menshakov
+    Copyright (C) 2015-2020  Vladimir Menshakov
 
     This library is free software; you can redistribute it and/or modify it
     under the terms of the GNU Lesser General Public License as published by
@@ -25,8 +25,12 @@
 #include "mtpobjectsmodel.h"
 #include "mtpstoragesmodel.h"
 #include "fileuploader.h"
+#include "devicesdialog.h"
 #include "utils.h"
-#include <usb/Context.h>
+#include <mtp/metadata/Library.h>
+#include <mtp/mtpz/TrustedApp.h>
+#include <mtp/ptp/Device.h>
+#include <mtp/ptp/Session.h>
 #include <mtp/usb/TimeoutException.h>
 #include <mtp/usb/DeviceBusyException.h>
 #include <mtp/usb/DeviceNotFoundException.h>
@@ -37,20 +41,44 @@
 #include <QKeyEvent>
 #include <QFileDialog>
 #include <QSettings>
+#include <QToolButton>
+#include <QNetworkAccessManager>
+#include <QNetworkRequest>
+#include <QNetworkReply>
+#include <QUrl>
+#if QT_VERSION >= 0x050000
+#	include <QStandardPaths>
+#else
+#	include <QDesktopServices>
+#endif
 #include <QDir>
 
 MainWindow::MainWindow(QWidget *parent) :
 	QMainWindow(parent),
-	_ui(new Ui::MainWindow),
+	_ui(new Ui::MainWindow), _nam(),
 	_clipboard(QApplication::clipboard()),
 	_proxyModel(new QSortFilterProxyModel),
 	_storageModel(),
 	_objectModel(new MtpObjectsModel()),
-	_uploader(new FileUploader(_objectModel, this))
+	_uploader(new FileUploader(_objectModel, this)),
+	_resetDevice(false),
+	_networkReply()
 {
 	_ui->setupUi(this);
 	setWindowIcon(QIcon(":/android-file-transfer.png"));
 
+	QString theme;
+	{
+		auto value = palette().text().color().value();
+		//trying to guess if it's dark or light theme
+		theme = value > 128? "dark": "light";
+		qDebug() << "current text color value: " << value << ", guessed theme: " << theme;
+	}
+	_ui->actionBack->setIcon(QIcon(":/icons/" + theme + "/go-previous.svg"));
+	_ui->actionGoDown->setIcon(QIcon(":/icons/" + theme + "/go-next.svg"));
+	_ui->actionCreateDirectory->setIcon(QIcon(":/icons/" + theme + "/folder-new.svg"));
+	_ui->actionRefresh->setIcon(QIcon(":/icons/" + theme + "/view-refresh.svg"));
+
 	_ui->listView->setModel(_proxyModel);
 
 	_proxyModel->setSortCaseSensitivity(Qt::CaseInsensitive);
@@ -63,10 +91,12 @@ MainWindow::MainWindow(QWidget *parent) :
 	connect(_ui->listView, SIGNAL(doubleClicked(QModelIndex)), SLOT(onActivated(QModelIndex)));
 	connect(_ui->listView, SIGNAL(customContextMenuRequested(QPoint)), SLOT(showContextMenu(QPoint)));
 	connect(_ui->actionBack, SIGNAL(triggered()), SLOT(back()));
-	connect(_ui->actionGo_Down, SIGNAL(triggered()), SLOT(down()));
+	connect(_ui->actionGoDown, SIGNAL(triggered()), SLOT(down()));
 	connect(_ui->actionCreateDirectory, SIGNAL(triggered()), SLOT(createDirectory()));
 	connect(_ui->actionUploadDirectory, SIGNAL(triggered()), SLOT(uploadDirectories()));
-	connect(_ui->actionUpload_Album, SIGNAL(triggered()), SLOT(uploadAlbum()));
+	connect(_ui->actionUploadAlbum, SIGNAL(triggered()), SLOT(uploadAlbum()));
+	connect(_ui->actionImportMusic, SIGNAL(triggered()), SLOT(importMusic()));
+	connect(_ui->actionImportMusicFiles, SIGNAL(triggered()), SLOT(importMusicFiles()));
 	connect(_ui->actionUpload, SIGNAL(triggered()), SLOT(uploadFiles()));
 	connect(_ui->actionRename, SIGNAL(triggered()), SLOT(renameFile()));
 	connect(_ui->actionDownload, SIGNAL(triggered()), SLOT(downloadFiles()));
@@ -75,16 +105,21 @@ MainWindow::MainWindow(QWidget *parent) :
 	connect(_ui->actionRefresh, SIGNAL(triggered()), SLOT(refresh()));
 	connect(_ui->actionPaste, SIGNAL(triggered()), SLOT(pasteFromClipboard()));
 	connect(_ui->actionShowThumbnails, SIGNAL(triggered(bool)), SLOT(showThumbnails(bool)));
+	connect(_ui->actionRemoveCover, SIGNAL(triggered(bool)), SLOT(removeCover()));
+	connect(_ui->actionAttachCover, SIGNAL(triggered(bool)), SLOT(attachCover()));
 
-	connect(_objectModel, SIGNAL(onFilesDropped(QStringList)), SLOT(uploadFiles(QStringList)));
+	connect(_objectModel, SIGNAL(onFilesDropped(QStringList)), SLOT(onFilesDropped(QStringList)));
 	connect(_objectModel, SIGNAL(existingFileOverwrite(QString)), SLOT(confirmOverwrite(QString)), Qt::BlockingQueuedConnection);
 
 	connect(_clipboard, SIGNAL(dataChanged()), SLOT(validateClipboard()));
 	validateClipboard();
+	QToolButton * importMusic = dynamic_cast<QToolButton *>(_ui->mainToolBar->widgetForAction(_ui->actionImportMusic));
+	importMusic->setMenu(new QMenu(tr("Import Music")));
+	importMusic->menu()->addAction(_ui->actionImportMusicFiles);
 
 	//fixme: find out how to specify alternative in designer
 	_ui->actionBack->setShortcuts(_ui->actionBack->shortcuts() << QKeySequence("Alt+Up") << QKeySequence("Esc"));
-	_ui->actionGo_Down->setShortcuts(_ui->actionGo_Down->shortcuts() << QKeySequence("Alt+Down") << QKeySequence("Enter"));
+	_ui->actionGoDown->setShortcuts(_ui->actionGoDown->shortcuts() << QKeySequence("Alt+Down") << QKeySequence("Enter"));
 	_ui->actionCreateDirectory->setShortcuts(_ui->actionCreateDirectory->shortcuts() << QKeySequence("F7"));
 	_ui->actionRefresh->setShortcuts(_ui->actionRefresh->shortcuts() << QKeySequence("Ctrl+R"));
 	_ui->listView->setFocus();
@@ -119,86 +154,86 @@ void MainWindow::closeEvent(QCloseEvent *event)
 	QMainWindow::closeEvent(event);
 }
 
+void MainWindow::replyReadyRead()
+{
+	if (!_networkReply) {
+		qWarning() << "replyReadyRead called without network reply object";
+		return;
+	}
+
+	auto data = _networkReply->read(128 * 1024);
+	qDebug() << "read: " << data.size() << " bytes...";
+	_networkReplyBody.append(data);
+}
+
+void MainWindow::replyFinished(QNetworkReply * reply)
+{
+	_networkReply = NULL;
+	qDebug() << "got reply " << reply << " with status " << reply->error();
+	auto title = tr("MTPZ Keys Download");
+	if (reply->error() != 0)
+	{
+		QMessageBox::warning(this, title, tr("Could not download keys, please find the error below:\n\n%1\n\nPlease look for .mtpz-data file on the internet and manually install it to your home directory.").arg(reply->errorString()));
+		return;
+	}
+	reply->open(QIODevice::ReadOnly);
+	auto mtpzDataPath = getMtpzDataPath();
+	QFile destination(mtpzDataPath);
+	qDebug() << "writing to " << destination.fileName();
+	if (!destination.open(QIODevice::WriteOnly)) {
+		QMessageBox::warning(this, title, tr("Could not write keys to %1").arg(mtpzDataPath));
+		return;
+	}
+
+	if (destination.write(_networkReplyBody) == -1)
+		QMessageBox::warning(this, title, tr("Could not write keys, please find the error below:\n\n%1\n\nPlease look for .mtpz-data file on the internet and manually install it to your home directory.").arg(destination.errorString()));
+
+	destination.close();
+	reply->close();
+	reply->deleteLater();
+	try
+	{
+		_trustedApp = mtp::TrustedApp::Create(_session, toUtf8(mtpzDataPath));
+		if (!_trustedApp->KeysLoaded())
+			throw std::runtime_error("failed to load new keys");
+
+		qDebug() << "new keys loaded, authenticating...";
+
+		_trustedApp->Authenticate();
+
+		QMessageBox::information(this, title, tr("MTPZ keys have been installed to your system."));
+		tryCreateLibrary();
+	}
+	catch (const std::exception & ex)
+	{
+		qWarning() << "failed to recreate session: " << ex.what();
+		QMessageBox::warning(this, title, tr("Your MTPZ keys failed to install or load.\n\nPlease restart the application to try again.\n\nException: %1").arg(ex.what()));
+	}
+
+}
+
+QString MainWindow::getMtpzDataPath()
+{
+#if QT_VERSION >= 0x050000
+	auto path = QStandardPaths::writableLocation(QStandardPaths::HomeLocation);
+#else
+	auto path = QDesktopServices::storageLocation(QDesktopServices::HomeLocation);
+#endif
+	path += "/.mtpz-data";
+	return path;
+}
+
 bool MainWindow::reconnectToDevice()
 {
 	_session.reset();
 	_device.reset();
-	bool claimInterface = true;
-	bool resetDevice = false;
 
-	mtp::usb::ContextPtr ctx(new mtp::usb::Context);
-
-	auto devices = ctx->GetDevices();
-	for (auto desc = devices.begin(); desc != devices.end();)
 	{
-		qDebug("probing device...");
-		try
-		{
-			_device = mtp::Device::Open(ctx, *desc, claimInterface, resetDevice);
-			++desc;
-		}
-		catch(const mtp::usb::DeviceBusyException &ex)
-		{
-			bool canKill = !ex.Processes.empty();
-			QString processList;
-			for (auto & desc : ex.Processes)
-			{
-				processList += QString("%1 (pid: %2)\n").arg(fromUtf8(desc.Name)).arg(desc.Id);
-			}
-			if (canKill)
-				processList = tr("The following processes are keeping file descriptors for your device:\n") + processList;
-
-			QMessageBox dialog(
-				QMessageBox::Warning,
-				tr("Device is busy"),
-				tr("Device is busy, maybe another process is using it.\n\n") +
-				processList +
-				tr("Close other MTP applications and restart Android File Transfer.\n"
-				"\nPress Abort to kill them or Ignore to try next device."),
-				(canKill? QMessageBox::Abort: QMessageBox::StandardButton(0)) | QMessageBox::Ignore,
-				this
-			);
-
-			dialog.setDefaultButton(QMessageBox::Ignore);
-			dialog.setEscapeButton(QMessageBox::Ignore);
-			auto r = dialog.exec();
-
-			if ((r & QMessageBox::Abort) == QMessageBox::Abort)
-			{
-				qDebug("kill'em all");
-				ex.Kill();
-				qDebug("retrying..."); //do not increment desc, retry device
-			}
-			if ((r & QMessageBox::Ignore) == QMessageBox::Ignore)
-				++desc;
-		}
-		catch(const std::exception &ex)
-		{
-			qWarning("Device::Find failed: %s", ex.what());
-			QMessageBox dialog(
-				QMessageBox::Warning,
-				tr("Device::Find failed"),
-				tr(
-					"MTP device could not be opened at the moment\n\nFailure: %1\n"
-					"Press Reset to reset them or Ignore to try next device.").arg(fromUtf8(ex.what())),
-				QMessageBox::Reset | QMessageBox::Ignore,
-				this
-			);
-			dialog.setDefaultButton(QMessageBox::Ignore);
-			dialog.setEscapeButton(QMessageBox::Ignore);
-			auto r = dialog.exec();
-
-			if ((r & QMessageBox::Reset) == QMessageBox::Reset)
-			{
-				qDebug("retry with reset...");
-				resetDevice = true;
-			}
-			if ((r & QMessageBox::Ignore) == QMessageBox::Ignore)
-				++desc;
-		}
-
-		if (_device)
-			break;
+		DevicesDialog dialog(_resetDevice, this);
+		int r = dialog.exec();
+		if (r == QDialog::Rejected)
+			return false;
+		_device = dialog.getDevice();
 	}
 
 	if (!_device)
@@ -217,6 +252,50 @@ bool MainWindow::reconnectToDevice()
 			_session = _device->OpenSession(1);
 			mtp::msg::DeviceInfo di = _session->GetDeviceInfo();
 			qDebug() << "device info" << fromUtf8(di.Manufacturer) << " " << fromUtf8(di.Model);
+			auto path = getMtpzDataPath();
+			qDebug() << "mtpz-data path: " << path;
+			_trustedApp = mtp::TrustedApp::Create(_session, toUtf8(path));
+			if (_trustedApp && !_trustedApp->KeysLoaded()) {
+				QString title = tr("MTPZ Keys are Missing");
+				QString header = tr(
+						"It seems your computer is missing an important bit for talking with MTPZ device: "
+						"private keys and certificate material from Microsoft.\n"
+						"This means that you can't download and upload files from/to such devices.\n\n"
+						"Microsoft could have released that key material and documentation for MTPZ devices "
+						"as they are not interested in those anymore.\n\n"
+						"Because of the legal risks we can't bundle those keys, even though in some countries it's lawful to modify things to make them working again, "
+						"just because you own it.\n\n"
+				);
+#ifdef MTPZ_DATA_SOURCE
+				QMessageBox downloadKeys(QMessageBox::Question,
+					title,
+					header + tr(
+						"Alternatively I (as an app) can offer you to download keys from the Internet.\n"
+						"Can I download keys for you?\n\n"
+
+						"(Please press Yes only if all of the above is legal in your country or you just don't care)."
+					),
+					QMessageBox::Yes | QMessageBox::No
+				);
+				int r = downloadKeys.exec();
+				if (r & QMessageBox::Yes) {
+					qDebug() << "downloading keys";
+					if (!_nam) {
+						_nam = new QNetworkAccessManager(this);
+						connect(_nam, SIGNAL(finished(QNetworkReply*)), this, SLOT(replyFinished(QNetworkReply*)));
+					}
+					_networkReply = _nam->get(QNetworkRequest(QUrl(MTPZ_DATA_SOURCE)));
+					connect(_networkReply, SIGNAL(readyRead()), this, SLOT(replyReadyRead()));
+				}
+#else
+				QMessageBox downloadKeys(QMessageBox::Warning,
+					title,
+					header + tr(
+						"You can look for .mtpz-data file on the internet, download it and place it in your home directory."
+					));
+				downloadKeys.exec();
+#endif
+			}
 			break;
 		}
 		catch(const mtp::usb::TimeoutException &ex)
@@ -295,6 +374,54 @@ void MainWindow::showEvent(QShowEvent *)
 		onStorageChanged(_ui->storageList->currentIndex());
 		qDebug() << "session opened, starting";
 		_proxyModel->setSourceModel(_objectModel);
+		if (_trustedApp && _trustedApp->KeysLoaded())
+		{
+			qDebug() << "keys loaded, authenticated";
+			_trustedApp->Authenticate();
+		}
+
+		tryCreateLibrary();
+
+		bool canUpload = _session->GetDeviceInfo().Supports(mtp::OperationCode::SendObjectInfo);
+		_ui->actionCreateDirectory->setVisible(canUpload);
+		_ui->actionUploadAlbum->setVisible(canUpload);
+		_ui->actionUploadDirectory->setVisible(canUpload);
+		_ui->actionUpload->setVisible(canUpload);
+	}
+}
+
+void MainWindow::tryCreateLibrary()
+{
+	if (_uploader->library())
+		return;
+
+	_ui->actionUploadAlbum->setVisible(true);
+	_ui->actionImportMusic->setVisible(false);
+
+	if (mtp::Library::Supported(_session)) {
+		ProgressDialog progressDialog(this, false);
+		progressDialog.setWindowTitle(tr("Loading Media Library"));
+		progressDialog.setModal(true);
+		progressDialog.setValue(0);
+
+		connect(_uploader, SIGNAL(uploadProgress(float)), &progressDialog, SLOT(setValue(float)));
+		connect(_uploader, SIGNAL(uploadSpeed(qint64)), &progressDialog, SLOT(setSpeed(qint64)));
+		connect(_uploader, SIGNAL(uploadStarted(QString)), &progressDialog, SLOT(setFilename(QString)));
+		connect(_uploader, SIGNAL(finished()), &progressDialog, SLOT(accept()));
+		connect(&progressDialog, SIGNAL(abort()), &progressDialog, SLOT(reject()));
+
+		try
+		{
+			_uploader->tryCreateLibrary();
+			_ui->actionUploadAlbum->setVisible(false);
+			_ui->actionImportMusic->setVisible(true);
+		}
+		catch (const std::exception & ex)
+		{
+			qWarning() << "importing music disabled: " << ex.what();
+		}
+
+		progressDialog.exec();
 	}
 }
 
@@ -340,7 +467,7 @@ void MainWindow::updateActionsState()
 	_ui->actionDelete->setEnabled(!rows.empty());
 	_ui->actionDownload->setEnabled(!rows.empty());
 	_ui->actionRename->setEnabled(rows.size() == 1);
-	_ui->actionGo_Down->setEnabled(rows.size() == 1);
+	_ui->actionGoDown->setEnabled(rows.size() == 1);
 	_ui->actionBack->setEnabled(!_history.empty());
 
 	QStringList statusList;
@@ -417,6 +544,38 @@ void MainWindow::showContextMenu ( const QPoint & pos )
 	menu.addAction(_ui->actionRename);
 	menu.addAction(_ui->actionDownload);
 	menu.addAction(_ui->actionDelete);
+
+	bool showRSMenu = false;
+
+	std::unordered_set<mtp::ObjectFormat> visited;
+	for(QModelIndex row : rows)
+	{
+		row = mapIndex(row);
+		auto id = _objectModel->objectIdAt(row.row());
+		try {
+			mtp::ObjectFormat format = static_cast<mtp::ObjectFormat>(_session->GetObjectIntegerProperty(id, mtp::ObjectProperty::ObjectFormat));
+			if (visited.find(format) != visited.end())
+				continue;
+
+			visited.insert(format);
+			auto supportedProperties = _session->GetObjectPropertiesSupported(format);
+			auto & properties = supportedProperties.ObjectPropertyCodes;
+			auto it = std::find(properties.begin(), properties.end(), mtp::ObjectProperty::RepresentativeSampleData);
+			if (it != properties.end()) {
+				showRSMenu = true;
+				break;
+			}
+		} catch (const std::exception & ex) {
+			qWarning() << "checking representative sample failed";
+		}
+	}
+	if (showRSMenu) {
+		menu.addSeparator();
+		if (rows.size() == 1)
+			menu.addAction(_ui->actionAttachCover);
+		menu.addAction(_ui->actionRemoveCover);
+	}
+
 	menu.exec(_ui->listView->mapToGlobal(pos));
 }
 
@@ -611,30 +770,66 @@ void MainWindow::uploadAlbum()
 	}
 }
 
-namespace
+void MainWindow::importMusic()
+{ importMusic(true); }
+
+void MainWindow::importMusicFiles()
+{ importMusic(false); }
+
+void MainWindow::importMusic(bool directoryMode)
 {
-	int GetScore(const QString &str_)
+	qDebug() << "import music, directory mode: " << directoryMode;
+
+	QFileDialog d(this);
+	QSettings settings;
 	{
-		QString str = str_.toLower();
-		int score = 0;
-		if (str.contains("art"))
-			score += 1;
-		if (str.contains("album"))
-			score += 1;
-		if (str.contains("cover"))
-			score += 2;
-		if (str.contains("large"))
-			score += 2;
-		if (str.contains("small"))
-			score += 1;
-		if (str.contains("folder"))
-			score += 1;
-		return score;
+		QVariant ld = settings.value("the-latest-directory");
+		if (ld.isValid())
+			d.setDirectory(ld.toString());
 	}
 
+	d.setAcceptMode(QFileDialog::AcceptOpen);
+	d.setOption(QFileDialog::ReadOnly, true);
+	if (directoryMode) {
+		d.setFileMode(QFileDialog::Directory);
+		d.setOption(QFileDialog::ShowDirsOnly, true);
+	} else {
+		d.setFileMode(QFileDialog::ExistingFiles);
+	}
+	restoreGeometry("import-music", d);
+	if (!d.exec())
+		return;
+
+	saveGeometry("import-music", d);
+	settings.setValue("the-latest-directory", d.directory().absolutePath());
+	QStringList selected = d.selectedFiles();
+	if (selected.isEmpty())
+		return;
+
+	for(const auto &path : selected)
+	{
+		importMusic(path);
+	}
+}
+
+void MainWindow::onFilesDropped(const QStringList &files)
+{
+	if (_uploader->library())
+	{
+		for(auto & file : files)
+		{
+			importMusic(file);
+		}
+	}
+	else
+		uploadFiles(files);
+}
+
+namespace
+{
 	bool HeuristicLess(const QString &s1, const QString &s2)
 	{
-		return GetScore(s1) > GetScore(s2);
+		return GetCoverScore(s1) > GetCoverScore(s2);
 	}
 }
 
@@ -652,7 +847,7 @@ void MainWindow::uploadAlbum(QString dirPath)
 	{
 		QStringList ext({"*.png", "*.jpg", "*.jpeg"});
 		covers = dir.entryList(ext, QDir::Files);
-		qSort(covers.begin(), covers.end(), &HeuristicLess);
+		std::sort(covers.begin(), covers.end(), &HeuristicLess);
 		qDebug() << "covers" << covers;
 		if (!covers.isEmpty())
 			cover = covers.front();
@@ -676,6 +871,24 @@ void MainWindow::uploadAlbum(QString dirPath)
 	uploadFiles(files);
 }
 
+void MainWindow::importMusic(QString path)
+{
+	qDebug() << "import music from " << path;
+
+	ProgressDialog progressDialog(this);
+	progressDialog.setWindowTitle(tr("Importing Progress"));
+	progressDialog.show();
+
+	connect(_uploader, SIGNAL(uploadProgress(float)), &progressDialog, SLOT(setValue(float)));
+	connect(_uploader, SIGNAL(uploadSpeed(qint64)), &progressDialog, SLOT(setSpeed(qint64)));
+	connect(_uploader, SIGNAL(uploadStarted(QString)), &progressDialog, SLOT(setFilename(QString)));
+	connect(_uploader, SIGNAL(finished()), &progressDialog, SLOT(accept()));
+	connect(&progressDialog, SIGNAL(abort()), _uploader, SLOT(abort()));
+	_uploader->importMusic(path);
+
+	progressDialog.exec();
+}
+
 void MainWindow::validateClipboard()
 {
 	QStringList files = _objectModel->extractMimeData(_clipboard->mimeData());
@@ -718,3 +931,72 @@ void MainWindow::showThumbnails(bool enable)
 	_ui->listView->setViewMode(enable? QListView::IconMode: QListView::ListMode);
 	_objectModel->enableThumbnail(enable, maxSize); //resets model/size hints, etc
 }
+
+void MainWindow::attachCover()
+{
+	QItemSelectionModel *selection =_ui->listView->selectionModel();
+	QModelIndexList rows = selection->selectedRows();
+	if (rows.empty())
+		return;
+
+	auto row = mapIndex(rows.at(0));
+	auto targetObjectId = _objectModel->objectIdAt(row.row());
+	qDebug() << "attaching cover to object " << targetObjectId.Id;
+
+	QFileDialog d(this);
+
+	QSettings settings;
+	{
+		QVariant ld = settings.value("the-latest-directory");
+		if (ld.isValid())
+			d.setDirectory(ld.toString());
+	}
+
+	d.setAcceptMode(QFileDialog::AcceptOpen);
+	d.setFileMode(QFileDialog::ExistingFile);
+	d.setOption(QFileDialog::ShowDirsOnly, false);
+	d.setOption(QFileDialog::ReadOnly, true);
+	restoreGeometry("upload-files", d);
+	if (!d.exec())
+		return;
+
+	saveGeometry("upload-files", d);
+	settings.setValue("the-latest-directory", d.directory().absolutePath());
+	auto files = d.selectedFiles();
+	if (files.empty())
+		return;
+
+	QFile file(files.at(0));
+	if (!file.open(QIODevice::ReadOnly)) {
+		QMessageBox::warning(this, tr("Attach Cover"), tr("Could not open selected file"));
+		return;
+	}
+	auto buffer = file.readAll();
+	qDebug() << "read " << buffer.size() << " bytes of cover file";
+	file.close();
+
+	mtp::ByteArray value(buffer.begin(), buffer.end());
+	try { _session->SetObjectPropertyAsArray(targetObjectId, mtp::ObjectProperty::RepresentativeSampleData, value); }
+	catch (const std::exception & ex)
+	{
+		QMessageBox::warning(this, tr("Attach Cover"), tr("Could not attach cover: %1").arg(ex.what()));
+	}
+}
+
+void MainWindow::removeCover()
+{
+	QItemSelectionModel *selection =_ui->listView->selectionModel();
+	QModelIndexList rows = selection->selectedRows();
+
+	mtp::ByteArray value;
+
+	for(QModelIndex row : rows)
+	{
+		row = mapIndex(row);
+		auto id = _objectModel->objectIdAt(row.row());
+		try
+		{ _session->SetObjectPropertyAsArray(id, mtp::ObjectProperty::RepresentativeSampleData, value); }
+		catch(const std::exception & ex)
+		{ qWarning() << "failed to remove cover:" << ex.what(); }
+	}
+}
