@@ -1,6 +1,6 @@
 /*
     This file is part of Android File Transfer For Linux.
-    Copyright (C) 2015-2018  Vladimir Menshakov
+    Copyright (C) 2015-2020  Vladimir Menshakov
 
     This library is free software; you can redistribute it and/or modify it
     under the terms of the GNU Lesser General Public License as published by
@@ -17,8 +17,8 @@
     Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
 
-#ifndef UTILS
-#define UTILS
+#ifndef AFTL_QT_UTILS_H
+#define AFTL_QT_UTILS_H
 
 #include <QString>
 #include <string>
@@ -34,5 +34,7 @@ inline std::string toUtf8(const QString &str)
 	return std::string(utf8.data(), utf8.size());
 }
 
+int GetCoverScore(const QString &str_);
+
 #endif // UTILS
 
